<?php
/**
 *
 * @category    payment gateway
 * @package     Tpaycom_Magento2.1
 * @author      tpay.com
 * @copyright   (https://tpay.com)
 */

namespace tpaycom\magento2cards\Service;

use Magento\Framework\App\ObjectManager;
use Magento\Framework\Exception\LocalizedException;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Model\Order\Invoice;
use Magento\Sales\Model\Order\Payment\Transaction;
use Magento\Sales\Model\Order\Payment\Transaction\BuilderInterface;
use Magento\Sales\Model\Order;
use tpaycom\magento2cards\Api\Sales\CardsOrderRepositoryInterface;
use tpaycom\magento2cards\lib\ResponseFields;

/**
 * Class TpayService
 *
 * @package tpaycom\magento2cards\Service
 */
class TpayService
{
    /**
     * @var CardsOrderRepositoryInterface
     */
    protected $orderRepository;

    protected $builder;

    private $objectManager;

    /**
     * TpayCards constructor.
     *
     * @param CardsOrderRepositoryInterface $orderRepository
     * @param BuilderInterface $builder
     */
    public function __construct(
        CardsOrderRepositoryInterface $orderRepository,
        BuilderInterface $builder
    ) {
        $this->orderRepository = $orderRepository;
        $this->builder = $builder;
        $this->objectManager = ObjectManager::getInstance();
    }

    /**
     * Change order state and notify user if needed
     *
     * @param int $orderId
     * @return Order
     */
    public function setOrderStatePendingPayment($orderId)
    {
        /** @var Order $order */
        $order = $this->orderRepository->getByIncrementId($orderId);

        $order->setTotalDue($order->getGrandTotal())
            ->setTotalPaid(0.00)
            ->setBaseTotalPaid(0.00)
            ->setBaseTotalDue($order->getBaseGrandTotal())
            ->setState(\Magento\Sales\Model\Order::STATE_PENDING_PAYMENT)
            ->addStatusToHistory(true);

        $order->save();

        return $order;
    }

    public function addCommentToHistory($orderId, $comment)
    {
        /** @var Order $order */
        $order = $this->orderRepository->getByIncrementId($orderId);
        $order->addStatusToHistory($order->getState(), $comment);
        $order->save();
    }

    /**
     * Return payment data
     *
     * @param int $orderId
     *
     * @return array
     */
    public function getPaymentData($orderId)
    {
        /** @var Order $order */
        $order = $this->orderRepository->getByIncrementId($orderId);

        return $order->getPayment()->getData();
    }

    /**
     * Validate order and set appropriate state
     *
     * @param int $orderId
     * @param array $validParams
     *
     * @return bool|Order
     */
    public function setOrderStatus($orderId, array $validParams)
    {
        /** @var Order $order */
        $order = $this->orderRepository->getByIncrementId($orderId);
        if (!$order->getId()) {
            return false;
        }
        $transactionDesc = $this->getTransactionDesc($validParams);
        $orderAmount = (double)number_format($order->getGrandTotal(), 2, '.', '');
        $emailNotify = false;

        if (!isset($validParams[ResponseFields::STATUS]) || $validParams[ResponseFields::STATUS] !== 'correct'
            || ((double)number_format($validParams[ResponseFields::AMOUNT], 2, '.', '') !== $orderAmount)
        ) {
            if ($order->getState() != Order::STATE_HOLDED) {
                $emailNotify = true;
            }
            $status = __('Payment has been declined: ') . '</br>' . $transactionDesc;
            $state = Order::STATE_HOLDED;
            $order->addStatusToHistory($state, $status, true);
        } else {
            if ($order->getState() != Order::STATE_PROCESSING) {
                $emailNotify = true;
            }
            $state = Order::STATE_PROCESSING;
            $this->setInvoice($order);
            $this->setTransaction($order, $validParams);
        }

        $order->setState($state);

        if ($emailNotify) {
            $order->setSendEmail(true);
        }
        $order->save();
        return $order;
    }

    /**
     * Get description for transaction
     *
     * @param array $validParams
     *
     * @return bool|string
     */
    protected function getTransactionDesc($validParams)
    {
        if (empty($validParams)) {
            return false;
        }
        if (isset ($validParams['err_desc'])) {
            return 'Payment error: ' . $validParams['err_desc'] . ', error code: ' . $validParams['err_code'];
        }
        $error = null;
        if ($validParams['status'] === 'declined') {
            $error = $validParams['reason'];
        }

        $transactionDesc = (is_null($error)) ? ' ' : ' Reason:  <b>' . $error . '</b>';
        $transactionDesc .= (isset($validParams[ResponseFields::TEST_MODE])) &&
        (int)$validParams[ResponseFields::TEST_MODE] === 1 ? '<b> TEST </b>' : ' ';
        return $transactionDesc;
    }

    /**
     * @param OrderInterface $order
     * @param array $validParams
     */
    private function setTransaction($order, $validParams)
    {
        $payment = $order->getPayment();

        if ($payment) {
            $payment->setLastTransId($validParams[ResponseFields::SALE_AUTH]);
            $payment->setTransactionId($validParams[ResponseFields::SALE_AUTH]);
            $payment->setAdditionalInformation(
                [\Magento\Sales\Model\Order\Payment\Transaction::RAW_DETAILS => (array)$validParams]
            );
            $trans = $this->builder;
            $transaction = $trans->setPayment($payment)
                ->setOrder($order)
                ->setTransactionId($validParams[ResponseFields::SALE_AUTH])
                ->setAdditionalInformation(
                    [\Magento\Sales\Model\Order\Payment\Transaction::RAW_DETAILS => (array)$validParams]
                )
                ->setFailSafe(true)
                //build method creates the transaction and returns the object
                ->build(\Magento\Sales\Model\Order\Payment\Transaction::TYPE_CAPTURE);

            $payment->setParentTransactionId(null)->registerCaptureNotification($order->getGrandTotal());
            $payment->save();
            $transaction->save();
            foreach ($order->getRelatedObjects() as $object) {
                $object->save();
            }
        }
    }

    /**
     * @param OrderInterface $order
     * @throws LocalizedException
     */
    private function setInvoice($order)
    {
        if ($order->canInvoice()) {
            // Create invoice for this order
            $invoice = $this->objectManager->create('Magento\Sales\Model\Service\InvoiceService')->prepareInvoice($order);

            // Make sure there is a qty on the invoice
            if (!$invoice->getTotalQty()) {
                throw new LocalizedException(
                    __('You can\'t create an invoice without products.')
                );
            }

        }
        // Register as invoice item
        $invoice->setRequestedCaptureCase(Invoice::NOT_CAPTURE);
        $invoice->register();
        $order->save();
        $transaction = $this->objectManager->create('Magento\Framework\DB\Transaction')
            ->addObject($invoice)
            ->addObject($invoice->getOrder());

        $transaction->save();

    }
}
